@@ -1,9 +1,9 @@
 # -*- test-case-name: twisted.trial.test.test_reporter -*-
 #
-# Copyright (c) 2001-2007 Twisted Matrix Laboratories.
+# Copyright (c) 2001-2008 Twisted Matrix Laboratories.
 # See LICENSE for details.
 #
-# Maintainer: Jonathan Lange <jml@twistedmatrix.com>
+# Maintainer: Jonathan Lange
 
 """
 Defines classes that handle the results of tests.
@@ -13,6 +13,7 @@ import sys, os
 import time
 import warnings
 
+from twisted.python.compat import set
 from twisted.python import reflect, log
 from twisted.python.components import proxyForInterface
 from twisted.python.failure import Failure
@@ -81,7 +82,8 @@ class TestResult(pyunit.TestResult, object):
         return error
 
     def startTest(self, test):
-        """This must be called before the given test is commenced.
+        """
+        This must be called before the given test is commenced.
 
         @type test: L{pyunit.TestCase}
         """
@@ -89,7 +91,8 @@ class TestResult(pyunit.TestResult, object):
         self._testStarted = self._getTime()
 
     def stopTest(self, test):
-        """This must be called after the given test is completed.
+        """
+        This must be called after the given test is completed.
 
         @type test: L{pyunit.TestCase}
         """
@@ -97,7 +100,8 @@ class TestResult(pyunit.TestResult, object):
         self._lastTime = self._getTime() - self._testStarted
 
     def addFailure(self, test, fail):
-        """Report a failed assertion for the given test.
+        """
+        Report a failed assertion for the given test.
 
         @type test: L{pyunit.TestCase}
         @type fail: L{Failure} or L{tuple}
@@ -105,10 +109,11 @@ class TestResult(pyunit.TestResult, object):
         self.failures.append((test, self._getFailure(fail)))
 
     def addError(self, test, error):
-        """Report an error that occurred while running the given test.
+        """
+        Report an error that occurred while running the given test.
 
         @type test: L{pyunit.TestCase}
-        @type fail: L{Failure} or L{tuple}
+        @type error: L{Failure} or L{tuple}
         """
         self.errors.append((test, self._getFailure(error)))
 
@@ -293,6 +298,20 @@ class _AdaptedReporter(TestResultDecorator):
 class Reporter(TestResult):
     """
     A basic L{TestResult} with support for writing to a stream.
+
+    @ivar _startTime: The time when the first test was started. It defaults to
+        C{None}, which means that no test was actually launched.
+    @type _startTime: C{float} or C{NoneType}
+
+    @ivar _warningCache: A C{set} of tuples of warning message (file, line,
+        text, category) which have already been written to the output stream
+        during the currently executing test.  This is used to avoid writing
+        duplicates of the same warning to the output stream.
+    @type _warningCache: C{set}
+
+    @ivar _publisher: The log publisher which will be observed for warning
+        events.
+    @type _publisher: L{LogPublisher} (or another type sufficiently similar)
     """
 
     implements(itrial.IReporter)
@@ -300,25 +319,51 @@ class Reporter(TestResult):
     _separator = '-' * 79
     _doubleSeparator = '=' * 79
 
-    def __init__(self, stream=sys.stdout, tbformat='default', realtime=False):
+    def __init__(self, stream=sys.stdout, tbformat='default', realtime=False,
+                 publisher=None):
         super(Reporter, self).__init__()
         self._stream = SafeStream(stream)
         self.tbformat = tbformat
         self.realtime = realtime
-        # The time when the first test was started.
         self._startTime = None
+        self._warningCache = set()
+
+        # Start observing log events so as to be able to report warnings.
+        self._publisher = publisher
+        if publisher is not None:
+            publisher.addObserver(self._observeWarnings)
+
+
+    def _observeWarnings(self, event):
+        """
+        Observe warning events and write them to C{self._stream}.
+
+        This method is a log observer which will be registered with
+        C{self._publisher.addObserver}.
+
+        @param event: A C{dict} from the logging system.  If it has a
+            C{'warning'} key, a logged warning will be extracted from it and
+            possibly written to C{self.stream}.
+        """
+        if 'warning' in event:
+            key = (event['filename'], event['lineno'],
+                   event['category'].split('.')[-1],
+                   str(event['warning']))
+            if key not in self._warningCache:
+                self._warningCache.add(key)
+                self._stream.write('%s:%s: %s: %s\n' % key)
 
 
     def stream(self):
         warnings.warn("stream is deprecated in Twisted 8.0.",
-                      category=DeprecationWarning, stacklevel=4)
+                      category=DeprecationWarning, stacklevel=2)
         return self._stream
     stream = property(stream)
 
 
     def separator(self):
         warnings.warn("separator is deprecated in Twisted 8.0.",
-                      category=DeprecationWarning, stacklevel=4)
+                      category=DeprecationWarning, stacklevel=2)
         return self._separator
     separator = property(separator)
 
@@ -326,13 +371,14 @@ class Reporter(TestResult):
     def startTest(self, test):
         """
         Called when a test begins to run. Records the time when it was first
-        called.
+        called and resets the warning cache.
 
         @param test: L{ITestCase}
         """
         super(Reporter, self).startTest(test)
         if self._startTime is None:
-            self._startTime = time.time()
+            self._startTime = self._getTime()
+        self._warningCache = set()
 
 
     def addFailure(self, test, fail):
@@ -581,6 +627,8 @@ class Reporter(TestResult):
         Expects that L{_printErrors}, L{_writeln}, L{_write}, L{_printSummary}
         and L{_separator} are all implemented.
         """
+        if self._publisher is not None:
+            self._publisher.removeObserver(self._observeWarnings)
         self._printErrors()
         self._writeln(self._separator)
         if self._startTime is not None:
@@ -611,7 +659,11 @@ class MinimalReporter(Reporter):
         %(num_failures) %(num_skips)'
         """
         numTests = self.testsRun
-        t = (self._startTime - self._getTime(), numTests, numTests,
+        if self._startTime is not None:
+            timing = self._getTime() - self._startTime
+        else:
+            timing = 0
+        t = (timing, numTests, numTests,
              len(self.errors), len(self.failures), len(self.skips))
         self._writeln(' '.join(map(str, t)))
 
@@ -734,24 +786,27 @@ class _AnsiColorizer(object):
     def __init__(self, stream):
         self.stream = stream
 
-    def supported(self):
+    def supported(cls, stream=sys.stdout):
         """
         A class method that returns True if the current platform supports
         coloring terminal output using this method. Returns False otherwise.
         """
-        # assuming stderr
-        # isatty() returns False when SSHd into Win32 machine
-        if 'CYGWIN' in os.environ:
-            return True
-        if not sys.stderr.isatty():
+        if not stream.isatty():
             return False # auto color only on TTYs
         try:
             import curses
-            curses.setupterm()
-            return curses.tigetnum("colors") > 2
-        except:
-            # guess false in case of error
+        except ImportError:
             return False
+        else:
+            try:
+                try:
+                    return curses.tigetnum("colors") > 2
+                except curses.error:
+                    curses.setupterm()
+                    return curses.tigetnum("colors") > 2
+            except:
+                # guess false in case of error
+                return False
     supported = classmethod(supported)
 
     def write(self, text, color):
@@ -789,7 +844,7 @@ class _Win32Colorizer(object):
             'white': red | green | blue | bold
             }
 
-    def supported(self):
+    def supported(cls, stream=sys.stdout):
         try:
             import win32console
             screenBuffer = win32console.GetStdHandle(
@@ -822,7 +877,7 @@ class _NullColorizer(object):
     def __init__(self, stream):
         self.stream = stream
 
-    def supported(self):
+    def supported(cls, stream=sys.stdout):
         return True
     supported = classmethod(supported)
 
@@ -850,11 +905,11 @@ class TreeReporter(Reporter):
     TODONE = 'red'
     SUCCESS = 'green'
 
-    def __init__(self, stream=sys.stdout, tbformat='default', realtime=False):
-        super(TreeReporter, self).__init__(stream, tbformat, realtime)
+    def __init__(self, stream=sys.stdout, *args, **kwargs):
+        super(TreeReporter, self).__init__(stream, *args, **kwargs)
         self._lastTest = []
         for colorizer in [_Win32Colorizer, _AnsiColorizer, _NullColorizer]:
-            if colorizer.supported():
+            if colorizer.supported(stream):
                 self._colorizer = colorizer(stream)
                 break
 
